"""Common helpers uses for all boundary operators."""

from bempp_cl.api.assembly import assembler as _assembler
from bempp_cl.api.assembly import boundary_operator as _boundary_operator
from bempp_cl.api.assembly import blocked_operator as _blocked_operator
import numpy as _np


# pylint: disable=too-many-arguments
def create_operator(
    identifier,
    domain,
    range_,
    dual_to_range,
    parameters,
    assembler,
    operator_options,
    kernel_type,
    assembly_type,
    device_interface,
    precision,
    is_complex,
):
    """Create a generic operator."""
    from bempp_cl.api.operators import OperatorDescriptor
    import bempp_cl.api

    if precision is None:
        precision = bempp_cl.api.DEFAULT_PRECISION

    if assembler != "only_singular_part":
        # Add operator for singular part assembly. Needed for certain fast methods.
        singular_part = create_operator(
            identifier,
            domain,
            range_,
            dual_to_range,
            parameters,
            "only_singular_part",
            operator_options,
            kernel_type,
            assembly_type,
            device_interface,
            precision,
            is_complex,
        )
    else:
        singular_part = None

    assembler = _assembler.AssemblerInterface(domain, dual_to_range, assembler, device_interface, precision, parameters)

    kernel_dimension = 1

    descriptor = OperatorDescriptor(
        identifier,
        operator_options,
        kernel_type,
        assembly_type,
        precision,
        is_complex,
        singular_part,
        kernel_dimension,
    )
    return _boundary_operator.BoundaryOperatorWithAssembler(domain, range_, dual_to_range, assembler, descriptor)


def create_multitrace_operator(
    identifier,
    domain,
    range_,
    dual_to_range,
    parameters,
    assembler,
    operator_options,
    multitrace_kernel,
    singular_contribution,
    device_interface,
    precision,
):
    """Create a generic operator."""
    from bempp_cl.api.operators import MultitraceOperatorDescriptor
    import bempp_cl.api

    if device_interface is None:
        device_interface = bempp_cl.api.DEFAULT_DEVICE

    if precision is None:
        precision = bempp_cl.api.DEFAULT_PRECISION

    assembler = _assembler.AssemblerInterface(domain, dual_to_range, assembler, device_interface, precision, parameters)

    descriptor = MultitraceOperatorDescriptor(identifier, operator_options, multitrace_kernel, singular_contribution)
    return _blocked_operator.MultitraceOperatorFromAssembler(domain, range_, dual_to_range, assembler, descriptor)


def pade_coeffs(order, angle):
    """
    Calculate the coefficients of the Pade series expansion of the square root.

    Parameters
    ----------
    order : int
        The order of the expansion.
    angle : float
        The branch-cut angle of the expansion.

    Returns
    -------
    c_0 : complex
    a_j : numpy.ndarray[complex]
    b_j : numpy.ndarray[complex]
    r_0 : complex
        The coefficients of the Pade expansion of the square root.
    """
    idx = _np.arange(order) + 1
    sin_j = 2 / (2 * order + 1) * _np.sin(idx * _np.pi / (2 * order + 1)) ** 2
    cos_j = _np.cos(idx * _np.pi / (2 * order + 1)) ** 2
    z = _np.exp(-1j * angle) - 1
    z_j = 1 + cos_j * z

    c_0 = _np.exp(0.5j * angle) * (1 + _np.sum(sin_j * z / z_j))
    a_j = _np.exp(-0.5j * angle) * sin_j / z_j**2
    b_j = _np.exp(-1j * angle) * cos_j / z_j
    r_0 = c_0 + _np.sum(a_j / b_j)

    return c_0, a_j, b_j, r_0


def inv_sqrt_pade_coeffs(order, angle=_np.pi / 3):
    """
    Calculate the coefficients of the Pade series expansion of the inverse square root.

    Parameters
    ----------
    order : int
        The order of the expansion.
    angle : float
        The branch-cut angle of the expansion (pi/3 is recommended).

    Returns
    -------
    s_j : numpy.ndarray[complex]
    r_j : numpy.ndarray[complex]
        The coefficients of the Pade expansion of the inverse square root.
    """
    idx = _np.arange(order)
    tan_j = _np.tan(0.5 * _np.pi * (0.5 + idx) / order) ** 2
    d_j = 1 + tan_j
    c_j = d_j / order

    r_j = _np.exp(1j * angle * 0.5) * c_j
    s_j = 1 + _np.exp(1j * angle) * (d_j - 1)

    return r_j, s_j



"""Interfaces to Laplace operators."""

from bempp_cl.api.operators.boundary import common as _common


def identity(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    device_interface=None,
    precision=None,
):
    """Assemble the L^2 identity operator."""
    return _common.create_operator(
        "l2_identity",
        domain,
        range_,
        dual_to_range,
        parameters,
        "sparse",
        [],
        "l2_identity",
        "default_sparse",
        device_interface,
        precision,
        False,
    )


def _vector_grad_product(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    device_interface=None,
    precision=None,
):
    """Assemble the inner product between a vector field and grad of a P1 basis function."""
    return _common.create_operator(
        "_vector_grad_product",
        domain,
        range_,
        dual_to_range,
        parameters,
        "sparse",
        [],
        "_vector_grad_product",
        "default_sparse",
        device_interface,
        precision,
        False,
    )


def _curl_curl_product(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    device_interface=None,
    precision=None,
):
    """Assemble inner product between the surface curl of 2 SNC basis functions."""
    return _common.create_operator(
        "_curl_curl_product",
        domain,
        range_,
        dual_to_range,
        parameters,
        "sparse",
        [],
        "_curl_curl_product",
        "default_sparse",
        device_interface,
        precision,
        False,
    )


def multitrace_identity(multitrace_operator, parameters=None, device_interface=None, precision=None):
    """
    Create a multitrace identity operator.

    Parameters
    ----------
    multitrace_operator : Bempp Operator
        A 2 x 2 multitrace operator object whose spaces are used to define
        the identity operator.

    Output
    ------
    A block-diagonal multitrace identity operator.

    """
    from bempp_cl.api.assembly.blocked_operator import BlockedOperator

    domain0, domain1 = multitrace_operator.domain_spaces
    dual_to_range0, dual_to_range1 = multitrace_operator.dual_to_range_spaces
    range0, range1 = multitrace_operator.range_spaces

    blocked_operator = BlockedOperator(2, 2)

    blocked_operator[0, 0] = identity(domain0, range0, dual_to_range0)
    blocked_operator[1, 1] = identity(domain1, range1, dual_to_range1)

    return blocked_operator


def mte_operators(domains_, ranges_, dual_to_ranges_, kappa):
    """
    Create basic sparse operators to assemble Pade approximate MtE operators.

    Parameters
    ----------
    domains_
        Domain spaces
    ranges_
        Range spaces
    dual_to_ranges_
        Dual spaces

    Output
    ------
    Basic sparse operators to assemble Pade approximate MtE operators.

    IP
        Identity operator built with piecewise linear basis functions
    IC
        Identity operator built with Hcurl conforming basis functions
    N
        Scaled curlcurl operator
    L
        Grad u dot v operator
    LT
        Transpose of L

    """
    IP = identity(domains_[1], ranges_[1], dual_to_ranges_[1])
    IC = identity(domains_[0], ranges_[0], dual_to_ranges_[0])
    N = (1.0 / kappa) ** 2 * _curl_curl_product(domains_[0], ranges_[0], dual_to_ranges_[0])
    LT = _vector_grad_product(domains_[1], ranges_[0], dual_to_ranges_[0])
    L = LT._transpose(LT._domain)
    return IP, IC, N, LT, L


def lambda_1(mte_operators, beta, kappa_eps):
    """Create and return block Pade approximate operator to Lambda1 = (I+Delta)^(1/2).

    Parameters
    ----------
    mte_operators
        Basic sparse Pade approximate MtE operators.
    beta
        Array containing the division between Pade coefficients: A_j/B_j
    kappa
        Damped wavenumber kappa_eps = kappa + i eps

    Output
    ------
    Block operator that approximates (I+Delta)^(1/2)
    """
    from scipy.sparse import bmat
    from bempp_cl.api.assembly.discrete_boundary_operator import InverseSparseDiscreteBoundaryOperator

    IP, IC, N, LT, L = mte_operators
    return InverseSparseDiscreteBoundaryOperator(
        bmat(
            [
                [(IC - beta * N).weak_form().to_sparse(), (beta * LT).weak_form().to_sparse()],
                [L.weak_form().to_sparse(), kappa_eps**2 * IP.weak_form().to_sparse()],
            ],
            "csc",
        )
    )


def lambda_2(mte_operators):
    """Create and return Lambda2 = (I-curlcurl) operator.

    Parameters
    ----------
    mte_operators

    Output
    ------
    IC-N = (I-curlcurl)

    """
    IP, IC, N, LT, L = mte_operators
    return IC - N


def sigma_identity(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    device_interface=None,
    precision=None,
):
    """
    Evaluate the sigma identity operator.

    For Galerkin methods this operator is equivalent to .5 * identity. For
    collocation methods the value may differ from .5 on piecewise smooth
    domains.
    """
    from bempp_cl.api.utils.helpers import assign_parameters

    parameters = assign_parameters(parameters)

    if parameters.assembly.discretization_type == "galerkin":
        return 0.5 * identity(
            domain,
            range_,
            dual_to_range,
            parameters=parameters,
            device_interface=device_interface,
            precision=precision,
        )
    elif parameters.assembly.discretization_type == "collocation":
        raise ValueError("Not yet implemented.")


def laplace_beltrami(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    device_interface=None,
    precision=None,
):
    """Assemble the negative Laplace-Beltrami operator."""
    if domain.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Domain shapeset must be of type 'p1_discontinuous'.")

    if dual_to_range.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Dual to range shapeset must be of type 'p1_discontinuous'.")

    return _common.create_operator(
        "laplace_beltrami",
        domain,
        range_,
        dual_to_range,
        parameters,
        "sparse",
        [],
        "laplace_beltrami",
        "default_sparse",
        device_interface,
        precision,
        False,
    )



"""Duffy transformation rules for singular integration for Galerkin integrals."""

import numpy as _np


def number_of_quadrature_points(order, adjacency):
    """Return the number of quadrature points for given adjacency.

    Possible cases for adjacency are "coincident", "edge_adjacent"
    or "vertex_adjacent".

    """
    if adjacency == "coincident":
        npoints = 6 * order**4
    elif adjacency == "edge_adjacent":
        npoints = 5 * order**4
    elif adjacency == "vertex_adjacent":
        npoints = 2 * order**4
    else:
        raise ValueError("adjacency must be one of 'coincident', " + "'edge_adjacent', 'vertex_adjacent'")

    return npoints


def rule(order, adjacency):
    """
    Create a singular quadrature rule.

    Possible cases for adjacency are "coincident",
    "edge_adjacent" or "vertex_adjacent".

    """
    from .gauss import rule as gauss_rule

    if adjacency not in ["coincident", "edge_adjacent", "vertex_adjacent"]:
        raise ValueError("Unknown adjacency.")

    xreg, wreg = gauss_rule(order)
    number_of_1d_points = len(wreg)
    # Create tensor Gauss points

    tensor_points = _np.empty((2, number_of_1d_points**2), dtype="float64")
    tensor_weights = _np.empty(number_of_1d_points**2, dtype="float64")

    for i in range(number_of_1d_points):
        for j in range(number_of_1d_points):
            tensor_points[0, i * number_of_1d_points + j] = xreg[j]
            tensor_points[1, i * number_of_1d_points + j] = xreg[i]
            tensor_weights[i * number_of_1d_points + j] = wreg[i] * wreg[j]

    number_of_reg_points = number_of_1d_points**2

    if adjacency == "coincident":
        number_of_points = 6 * number_of_reg_points**2
        points_test = _np.empty((2, number_of_points), dtype="float64", order="F")
        points_trial = _np.empty((2, number_of_points), dtype="float64", order="F")
        weights = _np.empty(number_of_points, dtype="float64")

        index = 0
        for test_ind in range(number_of_reg_points):
            for trial_ind in range(number_of_reg_points):
                ptest = tensor_points[:, test_ind]
                ptrial = tensor_points[:, trial_ind]

                xsi = ptest[0]
                eta1 = ptest[1]
                eta2 = ptrial[0]
                eta3 = ptrial[1]

                eta123 = eta1 * eta2 * eta3
                eta12 = eta1 * eta2

                weight = tensor_weights[test_ind] * tensor_weights[trial_ind] * xsi * xsi * xsi * eta1 * eta1 * eta2

                # Region 1
                points_test[0, index] = xsi
                points_test[1, index] = xsi * (1.0 - eta1 + eta12)
                points_trial[0, index] = xsi * (1.0 - eta123)
                points_trial[1, index] = xsi * (1.0 - eta1)
                weights[index] = weight
                index += 1

                # Region 2
                points_test[0, index] = xsi * (1.0 - eta123)
                points_test[1, index] = xsi * (1.0 - eta1)
                points_trial[0, index] = xsi
                points_trial[1, index] = xsi * (1.0 - eta1 + eta12)
                weights[index] = weight
                index += 1

                # Region 3
                points_test[0, index] = xsi
                points_test[1, index] = xsi * (eta1 - eta12 + eta123)
                points_trial[0, index] = xsi * (1.0 - eta12)
                points_trial[1, index] = xsi * (eta1 - eta12)
                weights[index] = weight
                index += 1

                # Region 4
                points_test[0, index] = xsi * (1.0 - eta12)
                points_test[1, index] = xsi * (eta1 - eta12)
                points_trial[0, index] = xsi
                points_trial[1, index] = xsi * (eta1 - eta12 + eta123)
                weights[index] = weight
                index += 1

                # Region 5
                points_test[0, index] = xsi * (1.0 - eta123)
                points_test[1, index] = xsi * (eta1 - eta123)
                points_trial[0, index] = xsi
                points_trial[1, index] = xsi * (eta1 - eta12)
                weights[index] = weight
                index += 1

                # Region 6
                points_test[0, index] = xsi
                points_test[1, index] = xsi * (eta1 - eta12)
                points_trial[0, index] = xsi * (1.0 - eta123)
                points_trial[1, index] = xsi * (eta1 - eta123)
                weights[index] = weight
                index += 1

    if adjacency == "edge_adjacent":
        number_of_points = 5 * number_of_reg_points**2
        points_test = _np.empty((2, number_of_points), dtype="float64", order="F")
        points_trial = _np.empty((2, number_of_points), dtype="float64", order="F")
        weights = _np.empty(number_of_points, dtype="float64")

        index = 0
        for test_ind in range(number_of_reg_points):
            for trial_ind in range(number_of_reg_points):
                ptest = tensor_points[:, test_ind]
                ptrial = tensor_points[:, trial_ind]

                xsi = ptest[0]
                eta1 = ptest[1]
                eta2 = ptrial[0]
                eta3 = ptrial[1]

                eta123 = eta1 * eta2 * eta3
                eta12 = eta1 * eta2

                weight = tensor_weights[test_ind] * tensor_weights[trial_ind] * xsi * xsi * xsi * eta1 * eta1

                # Region 1
                points_test[0, index] = xsi
                points_test[1, index] = xsi * eta1 * eta3
                points_trial[0, index] = xsi * (1.0 - eta12)
                points_trial[1, index] = xsi * eta1 * (1.0 - eta2)
                weights[index] = weight
                index += 1

                # Region 2
                points_test[0, index] = xsi
                points_test[1, index] = xsi * eta1
                points_trial[0, index] = xsi * (1.0 - eta123)
                points_trial[1, index] = xsi * eta1 * eta2 * (1 - eta3)
                weights[index] = weight * eta2
                index += 1

                # Region 3
                points_test[0, index] = xsi * (1.0 - eta12)
                points_test[1, index] = xsi * eta1 * (1.0 - eta2)
                points_trial[0, index] = xsi
                points_trial[1, index] = xsi * eta123
                weights[index] = weight * eta2
                index += 1

                # Region 4
                points_test[0, index] = xsi * (1.0 - eta123)
                points_test[1, index] = xsi * eta12 * (1.0 - eta3)
                points_trial[0, index] = xsi
                points_trial[1, index] = xsi * eta1
                weights[index] = weight * eta2
                index += 1

                # Region 5
                points_test[0, index] = xsi * (1.0 - eta123)
                points_test[1, index] = xsi * eta1 * (1.0 - eta2 * eta3)
                points_trial[0, index] = xsi
                points_trial[1, index] = xsi * eta12
                weights[index] = weight * eta2
                index += 1

    if adjacency == "vertex_adjacent":
        number_of_points = 2 * number_of_reg_points**2
        points_test = _np.empty((2, number_of_points), dtype="float64", order="F")
        points_trial = _np.empty((2, number_of_points), dtype="float64", order="F")
        weights = _np.empty(number_of_points, dtype="float64")

        index = 0
        for test_ind in range(number_of_reg_points):
            for trial_ind in range(number_of_reg_points):
                ptest = tensor_points[:, test_ind]
                ptrial = tensor_points[:, trial_ind]

                xsi = ptest[0]
                eta1 = ptest[1]
                eta2 = ptrial[0]
                eta3 = ptrial[1]

                weight = tensor_weights[test_ind] * tensor_weights[trial_ind] * xsi * xsi * xsi * eta2

                # Region 1
                points_test[0, index] = xsi
                points_test[1, index] = xsi * eta1
                points_trial[0, index] = xsi * eta2
                points_trial[1, index] = xsi * eta2 * eta3
                weights[index] = weight
                index += 1

                # Region 2
                points_test[0, index] = xsi * eta2
                points_test[1, index] = xsi * eta2 * eta3
                points_trial[0, index] = xsi
                points_trial[1, index] = xsi * eta1
                weights[index] = weight
                index += 1

    # Points above are for a different unit triangle than Bempp uses.
    # Fix this here.

    points_test[0, :] -= points_test[1, :]
    points_trial[0, :] -= points_trial[1, :]

    return points_test, points_trial, weights


def remap_points_shared_vertex(points, vertex_id):
    """
    Remap triangle points for vertex adjacency.

    By default the Duffy rules assume that the two triangles meet
    at vertex 0. This method transforms the Duffy integration
    points depending on which vertex they meet.
    """
    if vertex_id == 0:
        return points

    if vertex_id == 1:
        new_points = _np.zeros_like(points)
        new_points[0, :] = 1.0 - points[0, :] - points[1, :]
        new_points[1, :] = points[1, :]
        return new_points

    if vertex_id == 2:
        new_points = _np.zeros_like(points)
        new_points[0, :] = points[0, :]
        new_points[1, :] = 1.0 - points[0, :] - points[1, :]
        return new_points


def remap_points_shared_edge(points, shared_vertex1, shared_vertex2):
    """
    Remap triangle points for edge adjacency.

    By default the Duffy rules assume that the two triangles meet
    at edge 0. This method transforms the Duffy integration
    points depending on which edge they meet.
    """
    # Get the shared vertices

    v0 = shared_vertex1
    v1 = shared_vertex2

    # Create reference triangle
    ref_vertices = _np.zeros((2, 3), dtype="float64")
    ref_vertices[0, 1] = 1
    ref_vertices[1, 2] = 1

    new_vertices = _np.zeros((2, 3), dtype="float64")
    new_vertices[:, 0] = ref_vertices[:, v0]
    new_vertices[:, 1] = ref_vertices[:, v1]
    new_vertices[:, 2] = ref_vertices[:, 3 - v0 - v1]

    A = _np.zeros((2, 2), dtype="float64")
    A[:, 0] = new_vertices[:, 1] - new_vertices[:, 0]
    A[:, 1] = new_vertices[:, 2] - new_vertices[:, 0]

    # Linear transformation from reference points to
    # new points

    new_points = A.dot(points) + new_vertices[:, 0].reshape(2, 1)

    return _np.require(new_points, requirements=["F"])
